# Copyright 2023–2025 Ian Moore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

@dataclass(frozen=True)
class NetsEnum:
    ROLLUX: str = "rollux"
    ETH: str = "ethereum"
    ARB: str = "arbitrum"
    OP: str = "optimism"
    POLYGON: str = "polygon"
    LOCALHOST: str = "localhost"


# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

"""Raw log event data conversion helpers."""

import cachetools
from web3 import Web3
from typing import Union
from web3.contract import Contract
from web3.exceptions import BadFunctionCallOutput, ContractLogicError
from eth_typing import HexAddress
from eth_tester.exceptions import TransactionFailed
from ..data.token_details import TokenDetails
from ..abi.abi_load import ABILoad
from ..utils.base_utils import BaseUtils

_call_missing_exceptions = (TransactionFailed, BadFunctionCallOutput, ValueError, ContractLogicError)

DEFAULT_TOKEN_CACHE = cachetools.LRUCache(1024)

class TokenDetailError(Exception):
    """Cannot extract token details for an ERC-20 token for some reason."""

class Token:

    def create_token(
        self,
        web3: Web3,
        deployer: str,
        name: str,
        symbol: str,
        supply: int,
        decimals: int = 18,
    ) -> Contract:
        """Deploys a new ERC-20 token on local dev, testnet or mainnet.
    
        - Uses `ERC20Mock <https://github.com/sushiswap/sushiswap/blob/canary/contracts/mocks/ERC20Mock.sol>`_ contract for the deployment.
    
        - Waits until the transaction has completed
    
        Example:
    
        .. code-block::
    
            # Deploys an ERC-20 token where 100,000 tokens are allocated ato the deployer address
            token = create_token(web3, deployer, "Hentai books token", "HENTAI", 100_000 * 10**18)
            print(f"Deployed token contract address is {token.address}")
            print(f"Deployer account {deployer} has {token.functions.balanceOf(user_1).call() / 10**18} tokens")
    
        Find more examples in :ref:`tutorials` and unit testing source code.
    
        :param web3:
            Web3 instance
    
        :param deployer:
            Deployer account as 0x address.
    
            Make sure this account has enough ETH or native token to cover the gas cost.
    
        :param name: Token name
    
        :param symbol: Token symbol
    
        :param supply: Token starting supply as raw units.
    
            E.g. ``500 * 10**18`` to have 500 tokens minted to the deployer
            at the start.
    
        :param decimals: How many decimals ERC-20 token values have
    
        :return:
            Instance to a deployed Web3 contract.
        """
        return deploy_contract(web3, "ERC20MockDecimals.json", deployer, name, symbol, supply, decimals)
    
    
    def fetch_erc20_details(
        self,
        web3: Web3,
        token_address: Union[HexAddress, str],
        max_str_length: int = 256,
        raise_on_error=True,
        contract_name="ERC20MockDecimals.json",
        cache: cachetools.Cache | None = DEFAULT_TOKEN_CACHE,
        chain_id: int = None,
    ) -> TokenDetails:
        """Read token details from on-chain data.
    
        Connect to Web3 node and do RPC calls to extract the token info.
        We apply some sanitazation for incoming data, like length checks and removal of null bytes.
    
        The function should not raise an exception as long as the underlying node connection does not fail.
    
        Example:
    
        .. code-block:: python
    
            details = fetch_erc20_details(web3, token_address)
            assert details.name == "Hentai books token"
            assert details.decimals == 6
    
        :param web3:
            Web3 instance
    
        :param token_address:
            ERC-20 contract address:
    
        :param max_str_length:
            For input sanitisation
    
        :param raise_on_error:
            If set, raise `TokenDetailError` on any error instead of silently ignoring in and setting details to None.
    
        :param contract_name:
            Contract ABI file to use.
    
            The default is ``ERC20MockDecimals.json``. For USDC use ``centre/FiatToken.json``.
    
        :param cache:
            Use this cache for cache token detail calls.
    
            The main purpose is to easily reduce JSON-RPC API call count.
    
            By default, we use LRU cache of 1024 entries.
    
            Set to ``None`` to disable the cache.
    
            Instance of :py:class:`cachetools.Cache'.
            See `cachetools documentation for details <https://cachetools.readthedocs.io/en/latest/#cachetools.LRUCache>`__.
    
        :param chain_id:
            Chain id hint for the cache.
    
            If not given do ``eth_chainId`` RPC call to figure out.
    
        :return:
            Sanitised token info
        """
    
        if not chain_id:
            chain_id = web3.eth.chain_id
    
        erc_20 = ABILoad().get_deployed_contract(web3, contract_name, token_address)
        key = TokenDetails.generate_cache_key(chain_id, token_address)
    
        if cache is not None:
            cached = cache.get(key)
            if cached is not None:
                return TokenDetails(
                    erc_20,
                    cached["name"],
                    cached["symbol"],
                    cached["supply"],
                    cached["decimals"],
                )
        try:
            symbol = BaseUtils().sanitise_string(erc_20.functions.symbol().call()[0:max_str_length])
        except _call_missing_exceptions as e:
            if raise_on_error:
                raise TokenDetailError(f"Token {token_address} missing symbol") from e
            symbol = None
        except OverflowError:
            # OverflowError: Python int too large to convert to C ssize_t
            # Que?
            # Sai Stablecoin uses bytes32 instead of string for name and symbol information
            # https://etherscan.io/address/0x89d24a6b4ccb1b6faa2625fe562bdd9a23260359#readContract
            symbol = None
    
        try:
            name = BaseUtils().sanitise_string(erc_20.functions.name().call()[0:max_str_length])
        except _call_missing_exceptions as e:
            if raise_on_error:
                raise TokenDetailError(f"Token {token_address} missing name") from e
            name = None
        except OverflowError:
            # OverflowError: Python int too large to convert to C ssize_t
            # Que?
            # Sai Stablecoin uses bytes32 instead of string for name and symbol information
            # https://etherscan.io/address/0x89d24a6b4ccb1b6faa2625fe562bdd9a23260359#readContract
            name = None
    
        try:
            decimals = erc_20.functions.decimals().call()
        except _call_missing_exceptions as e:
            if raise_on_error:
                raise TokenDetailError(f"Token {token_address} missing decimals") from e
            decimals = 0
    
        try:
            supply = erc_20.functions.totalSupply().call()
        except _call_missing_exceptions as e:
            if raise_on_error:
                raise TokenDetailError(f"Token {token_address} missing totalSupply") from e
            supply = None
    
        token_details = TokenDetails(erc_20, name, symbol, supply, decimals)
        if cache is not None:
            cache[key] = {
                "name": name,
                "symbol": symbol,
                "supply": supply,
                "decimals": decimals,
            }
        return token_details
    
    
    def reset_default_token_cache(self):
        """Purge the cached token data.
    
        See :py:data:`DEFAULT_TOKEN_CACHE`
        """
        global DEFAULT_TOKEN_CACHE
        # Cache has a horrible API
        DEFAULT_TOKEN_CACHE.__dict__["_LRUCache__order"] = OrderedDict()
        DEFAULT_TOKEN_CACHE.__dict__["_Cache__currsize"] = 0
        DEFAULT_TOKEN_CACHE.__dict__["_Cache__data"] = dict()
 




# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

from typing import Optional, Protocol
from ..event.tools.log_context import LogContext

# For typing.Protocol see https://stackoverflow.com/questions/68472236/type-hint-for-callable-that-takes-kwargs
class ProgressUpdate(Protocol):
    """Informs any listener about the state of an event scan.

    Called before a new block is processed.

    Hook this up with `tqdm` for an interactive progress bar.
    """

    def __call__(
        self,
        current_block: int,
        start_block: int,
        end_block: int,
        chunk_size: int,
        total_events: int,
        last_timestamp: Optional[int],
        context: LogContext,
    ):
        """
        :param current_block:
            The block we are about to scan.
            After this scan, we have scanned `current_block + chunk_size` blocks

        :param start_block:
            The first block in our total scan range.

        :param end_block:
            The last block in our total scan range.

        :param chunk_size:
            What was the chunk size (can differ for the last scanned chunk)

        :param total_events:
            Total events picked up so far

        :param last_timestamp:
            UNIX timestamp of last picked up event (if any events picked up)

        :param context:
            Current context
        """


# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0


"""High performance EVM event reader.

For further reading see:

- `Ethereum JSON-RPC API spec <https://playground.open-rpc.org/?schemaUrl=https://raw.githubusercontent.com/ethereum/execution-apis/assembled-spec/openrpc.json&uiSchema%5BappBar%5D%5Bui:splitView%5D=false&uiSchema%5BappBar%5D%5Bui:input%5D=false&uiSchema%5BappBar%5D%5Bui:examplesDropdown%5D=false>`_

- `futureproof - - Bulletproof concurrent.futures for Python <https://github.com/yeraydiazdiaz/futureproof>`_

"""

import logging
from web3 import Web3
from web3.contract.contract import ContractEvent
from web3.datastructures import AttributeDict
from typing import Callable, Optional, Iterable, Dict, List
from eth_bloom import BloomFilter
from hexbytes import HexBytes

from ..tools.conversion import Conversion
from ..tools.log_context import LogContext
from ..tools.log_result import LogResult
from ...utils.progress_update import ProgressUpdate
from ...data.filter import Filter
from ...data.reorganization_monitor import ReorganizationMonitor

logger = logging.getLogger(__name__)

def extract_timestamps_json_rpc(
    web3: Web3,
    start_block: int,
    end_block: int,
) -> Dict[str, int]:
    """Get block timestamps from block headers.

    Use slow JSON-RPC block headers call to get this information.

    TODO: This is an old code path. This has been replaced by more robust
    :py:class:`ReorganisationMonitor` implementation.

    :return:
        block hash -> UNIX timestamp mapping
    """
    timestamps = {}
    convert = Conversion()

    logging.debug("Extracting timestamps for logs %d - %d", start_block, end_block)

    # Collect block timestamps from the headers
    for block_num in range(start_block, end_block + 1):
        raw_result = web3.manager.request_blocking("eth_getBlockByNumber", (hex(block_num), False))
        data_block_number = convert.convert_jsonrpc_value_to_int(raw_result["number"])
        assert data_block_number == block_num, "Blockchain node did not give us the block we want"
        timestamps[raw_result["hash"]] = convert.convert_jsonrpc_value_to_int(raw_result["timestamp"])

    return timestamps

class ReadEvents:

    def apply(
        self,
        web3: Web3,
        start_block: int,
        end_block: int,
        events: Optional[List[ContractEvent]] = None,
        notify: Optional[ProgressUpdate] = None,
        chunk_size: int = 100,
        context: Optional[LogContext] = None,
        extract_timestamps: Optional[Callable] = extract_timestamps_json_rpc,
        filter: Optional[Filter] = None,
        reorg_mon: Optional[ReorganizationMonitor] = None,
    ) -> Iterable[LogResult]:
        """Reads multiple events from the blockchain.
    
        Optimized to read multiple events from test blockchains.
    
        .. note ::
    
            For a much faster event reader check :py:class:`eth_defi.reader.multithread.MultithreadEventReader`.
            This implementation is mostly good with EVM test backends or very small block ranges.
    
        - Scans chains block by block
    
        - Returns events as a dict for optimal performance
    
        - Supports interactive progress bar
    
        - Reads all the events matching signature - any filtering must be done
          by the reader
    
        See `scripts/read-uniswap-v2-pairs-and-swaps.py` for a full example.
    
        Example:
    
        .. code-block:: python
    
            json_rpc_url = os.environ["JSON_RPC_URL"]
            web3 = Web3(HTTPProvider(json_rpc_url)
    
            web3.middleware_onion.clear()
    
            # Get contracts
            Factory = get_contract(web3, "sushi/UniswapV2Factory.json")
    
            start_block = 1
            end_block = web3.eth.block_number
    
            filter = Filter.create_filter(
                factory_address,
                [Factory.events.PairCreated],
            )
    
            # Read through all the events, all the chain, using a single threaded slow loop.
            # Only suitable for test EVM backends.
            pairs = []
            log: LogResult
            for log in read_events(
                web3,
                start_block,
                end_block,
                filter=filter,
                extract_timestamps=None,
            ):
                # Signature this
                #
                #  event PairCreated(address indexed token0, address indexed token1, address pair, uint);
                #
                # topic 0 = keccak(event signature)
                # topic 1 = token 0
                # topic 2 = token 1
                # argument 0 = pair
                # argument 1 = pair id
                #
                # log for EthereumTester backend is
                #
                # {'type': 'mined',
                #  'logIndex': 0,
                #  'transactionIndex': 0,
                #  'transactionHash': HexBytes('0x2cf4563f8c275e5b5d7a4e5496bfbaf15cc00d530f15f730ac4a0decbc01d963'),
                #  'blockHash': HexBytes('0x7c0c6363bc8f4eac452a37e45248a720ff09f330117cdfac67640d31d140dc38'),
                #  'blockNumber': 6,
                #  'address': '0xF2E246BB76DF876Cef8b38ae84130F4F55De395b',
                #  'data': HexBytes('0x00000000000000000000000068931307edcb44c3389c507dab8d5d64d242e58f0000000000000000000000000000000000000000000000000000000000000001'),
                #  'topics': [HexBytes('0x0d3648bd0f6ba80134a33ba9275ac585d9d315f0ad8355cddefde31afa28d0e9'),
                #   HexBytes('0x0000000000000000000000002946259e0334f33a064106302415ad3391bed384'),
                #   HexBytes('0x000000000000000000000000b9816fc57977d5a786e654c7cf76767be63b966e')],
                #  'context': None,
                #  'event': web3._utils.datatypes.PairCreated,
                #  'chunk_id': 1,
                #  'timestamp': None}
                #
                arguments = decode_data(log["data"])
                topics = log["topics"]
                token0 = convert_uint256_hex_string_to_address(topics[1])
                token1 = convert_uint256_hex_string_to_address(topics[2])
                pair_address = convert_uint256_bytes_to_address(arguments[0])
                pair_id = convert_int256_bytes_to_int(arguments[1])
    
                token0_details = fetch_erc20_details(web3, token0)
                token1_details = fetch_erc20_details(web3, token1)
    
    
        :param web3:
            Web3 instance
    
        :param events:
            List of Web3.py contract event classes to scan for.
    
            Pass this or filter.
    
        :param notify:
            Optional callback to be called before starting to scan each chunk
    
        :param start_block:
            First block to process (inclusive)
    
        :param end_block:
            Last block to process (inclusive)
    
        :param extract_timestamps:
            Override for different block timestamp extraction methods.
    
            This might need to use expensive`eth_getBlockByNumber` JSON-RPC API call.
            It will seriously slow down event reading.
            Set `extract_timestamps` to `None` to not get timestamps, but fast event lookups.
    
        :param chunk_size:
            How many blocks to scan in one eth_getLogs call
    
        :param context:
            Passed to the all generated logs
    
        :param filter:
            Pass a custom event filter for the readers
    
            Pass this or events.
    
        :param reorg_mon:
            If passed, use this instance to monitor and raise chain reorganisation exceptions.
    
        :return:
            Iterate over :py:class:`LogResult` instances for each event matched in
            the filter.
        """
    
        assert type(start_block) == int
        assert type(end_block) == int
    
        total_events = 0
    
        # TODO: retry middleware makes an exception
        # assert len(web3.middleware_onion) == 0, f"Must not have any Web3 middleware installed to slow down scan, has {web3.middleware_onion.middlewares}"
    
        # Construct our bloom filter
        if filter is None:
            assert events is not None, "Cannot pass both filter and events"
            filter = prepare_filter(events)
    
        last_timestamp = None
    
        for block_num in range(start_block, end_block + 1, chunk_size):
            last_of_chunk = min(end_block, block_num + chunk_size - 1)
    
            logger.debug("Extracting eth_getLogs from %d - %d", block_num, last_of_chunk)
    
            batch_events = 0
    
            # Stream the events
            for event in self.extract_events(
                web3,
                block_num,
                last_of_chunk,
                filter,
                context,
                extract_timestamps,
                reorg_mon,
            ):
                last_timestamp = event.get("timestamp")
                total_events += 1
                batch_events += 1
                yield event
    
            # Ping our master,
            # only when we have an event hit not to cause unnecessary block header fetches
            # TODO: Add argument notify always
            if notify is not None and batch_events:
                notify(block_num, start_block, end_block, chunk_size, total_events, last_timestamp, context)


    def extract_events(
        self,
        web3: Web3,
        start_block: int,
        end_block: int,
        filter: Filter,
        context: Optional[LogContext] = None,
        extract_timestamps: Optional[Callable] = extract_timestamps_json_rpc,
        reorg_mon: Optional[ReorganizationMonitor] = None,
    ) -> Iterable[LogResult]:
        """Perform eth_getLogs call over a block range.
    
        You should use :py:func:`read_events` unless you know the block range is something your node can handle.
    
        :param start_block:
            First block to process (inclusive)
    
        :param end_block:
            Last block to process (inclusive)
    
        :param filter:
            Internal filter used to match logs
    
        :param extract_timestamps:
            Method to get the block timestamps.
    
            This might need to use expensive`eth_getBlockByNumber` JSON-RPC API call.
            It will seriously slow down event reading.
            Set `extract_timestamps` to `None` to not get timestamps, but fast event lookups.
    
    
        :param context:
            Passed to the all generated logs
    
        :param reorg_mon:
            If passed, use this instance to monitor and raise chain reorganisation exceptions.
    
        :return:
            Iterable for the raw event data
        """
    
        if reorg_mon:
            assert extract_timestamps is None, "You cannot pass both reorg_mon and extract_timestamps"
    
        topics = list(filter.topics.keys())
    
        # https://www.quicknode.com/docs/ethereum/eth_getLogs
        # https://docs.alchemy.com/alchemy/guides/eth_getlogs
        filter_params = {
            "topics": [topics],  # JSON-RPC has totally braindead API to say how to do OR event lookup
            "fromBlock": hex(start_block),
            "toBlock": hex(end_block),
        }
    
        # Do the filtering by address.
        # eth_getLogs gets single address or JSON list of addresses
        if filter.contract_address:
            assert type(filter.contract_address) in (list, str), f"Got: {type(filter.contract_address)}"
            filter_params["address"] = filter.contract_address
    
        # logging.debug("Extracting logs %s", filter_params)
        # logging.info("Log range %d - %d", start_block, end_block)
    
        try:
            logs = web3.manager.request_blocking("eth_getLogs", (filter_params,))
        except Exception as e:
            block_count = end_block - start_block
            raise ReadingLogsFailed(f"eth_getLogs failed for {start_block:,} - {end_block:,} (total {block_count:,} with filter {filter}") from e

        convert = Conversion()
        
        if logs:
            if extract_timestamps is not None:
                timestamps = extract_timestamps(web3, start_block, end_block)
                if timestamps is None:
                    raise BadTimestampValueReturned("extract_timestamps returned None")
            else:
                timestamps = None
    
            for log in logs:
                block_hash = log["blockHash"]
                block_number = convert.convert_jsonrpc_value_to_int(log["blockNumber"])
                # Retrofit our information to the dict
                event_signature = log["topics"][0]
    
                if isinstance(log, AttributeDict):
                    # The following code is not going to work, because AttributeDict magic
                    raise RuntimeError("web3.py AttributeDict middleware detected. Please remove it with web3.middleware_onion.remove('attrdict') or from web3.provider.middleware list before attempting to read events")
    
                if isinstance(log["data"], HexBytes):
                    raise RuntimeError("web3.py pythonic middleware detected. Please remove it with web3.middleware_onion.remove('pythonic') before attempting to read events")
    
                log["context"] = context
    
                if type(event_signature) == HexBytes:
                    # Make sure we use lowercase string notation everywhere
                    event_signature = event_signature.hex()
    
                log["event"] = filter.topics[event_signature]
    
                # Can be hex string or integer (EthereumTester)
                log["blockNumber"] = convert.convert_jsonrpc_value_to_int(log["blockNumber"])
    
                # Used for debugging if we are getting bad data from node
                # or internally confused
                log["chunk_id"] = start_block
    
                if reorg_mon:
                    # Raises exception if chain tip has changed
                    timestamp = reorg_mon.check_block_reorg(block_number, block_hash)
                    assert timestamp is not None, f"Timestamp missing for block number {block_number}, hash {block_hash}. reorg known last block is: {reorg_mon.get_last_block_read()}"
                    log["timestamp"] = timestamp
                else:
                    if timestamps is not None:
                        try:
                            log["timestamp"] = timestamps[block_hash]
                            if type(log["timestamp"]) not in (int, float):
                                raise BadTimestampValueReturned(f"Timestamp was not int or float: {type(log['timestamp'])}: {type(log['timestamp'])}")
                        except KeyError as e:
                            # Reorg mon would handle this natively
                            raise TimestampNotFound(f"EVM event reader cannot match timestamp.\n" f"Timestamp missing for block number {block_number:,}, hash {block_hash}.\n" f" our timestamp table has {len(timestamps)} blocks.") from e
                    else:
                        # Not set, because reorg mon and timestamp extractor not provided,
                        # the caller must do the timestamp resolution themselves
                        log["timestamp"] = None
    
                yield log
    
    

    def prepare_filter(self, events: List[ContractEvent]) -> Filter:
        """Creates internal filter to match contract events."""
    
        # Construct our bloom filter
        bloom = BloomFilter()
        topics = {}
    
        for event in events:
            signatures = event.build_filter().topics
    
            for signature in signatures:
                topics[signature] = event
                # TODO: Confirm correct usage of bloom filter for topics
                bloom.add(bytes.fromhex(signature[2:]))
    
        filter = Filter(topics, bloom)
    
        return filter



# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0


import logging
from web3 import Web3
from hexbytes import HexBytes
from typing import Iterable
from ...data.reorganization_monitor import ReorganizationMonitor

from eth_defi.event_reader.block_header import BlockHeader

logger = logging.getLogger(__name__)

class JSONRPCReorganizationMonitor(ReorganizationMonitor):
    """Watch blockchain for reorgs using eth_getBlockByNumber JSON-RPC API.

    - Use expensive eth_getBlockByNumber call to download
      block hash and timestamp from Ethereum compatible node
    """

    def __init__(self, web3: Web3, **kwargs):
        super().__init__(**kwargs)
        self.web3 = web3

    def __repr__(self):
        return f"<JSONRPCReorganisationMonitor, last_block_read: {self.last_block_read}>"

    def get_last_block_live(self):
        return self.web3.eth.block_number

    def fetch_block_data(self, start_block, end_block) -> Iterable[BlockHeader]:
        total = end_block - start_block
        logger.debug(f"Fetching block headers and timestamps for logs {start_block:,} - {end_block:,}, total {total:,} blocks")
        web3 = self.web3

        # Collect block timestamps from the headers
        for block_num in range(start_block, end_block + 1):
            response_json = web3.manager._make_request("eth_getBlockByNumber", (hex(block_num), False))
            raw_result = response_json["result"]

            # Happens the chain tip and https://polygon-rpc.com/
            # - likely the request routed to different backend node
            if raw_result is None:
                logger.debug("Abnormally terminated at block %d, chain tip unstable?", block_num)
                break

            data_block_number = raw_result["number"]

            block_hash = raw_result["hash"]
            if isinstance(block_hash, HexBytes):
                # Web3.py middleware madness
                block_hash = block_hash.hex()

            if type(data_block_number) == str:
                # Real node
                assert int(raw_result["number"], 16) == block_num
                timestamp = int(raw_result["timestamp"], 16)
            else:
                # EthereumTester
                timestamp = raw_result["timestamp"]

            record = BlockHeader(block_num, block_hash, timestamp)
            logger.debug("Fetched block record: %s, total %d transactions", record, len(raw_result["transactions"]))
            yield record


# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

class ChainReorganizationDetected(Exception):
    block_number: int
    original_hash: str
    new_hash: str

    def __init__(self, block_number: int, original_hash: str, new_hash: str):
        self.block_number = block_number
        self.original_hash = original_hash
        self.new_hash = new_hash

        super().__init__(f"Block reorg detected at #{block_number:,}. Original hash: {original_hash}. New hash: {new_hash}")


# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

from typing import Union, Optional
from eth_typing import HexAddress
from ..abi.abi_load import ABILoad
from ..data.pair import PairDetails
from ..token.token import Token

class FetchPairDetails:

    def __init__(self):
        pass
    
    def apply(
        self,
        web3,
        pair_contact_address: Union[str, HexAddress],
        reverse_token_order: Optional[bool] = None,
        base_token_address: Optional[str] = None,
        quote_token_address: Optional[str] = None,
    ) -> PairDetails:
        """Get pair info for PancakeSwap, others.
    
        See also :py:class:`PairDetails`.
    
        :param web3:
            Web3 instance
    
        :param pair_contact_address:
            Smart contract address of trading pair
    
        :param reverse_token_order:
            Set the human readable token order.
    
            See :py:class`PairDetails` for more info.
    
        :param base_token_address:
            Automatically determine token order from addresses.
    
        :param quote_token_address:
            Automatically determine token order from addresses.
    
        """
    
        if base_token_address or quote_token_address:
            assert reverse_token_order is None, f"Give either (base_token_address, quote_token_address) or reverse_token_order"
            reverse_token_order = int(base_token_address, 16) > int(quote_token_address, 16)
    
        pool = ABILoad().get_deployed_contract(web3, "sushi/UniswapV2Pair.json", pair_contact_address)
        token0_address = pool.functions.token0().call()
        token1_address = pool.functions.token1().call()
    
        token0 = Token().fetch_erc20_details(web3, token0_address)
        token1 = Token().fetch_erc20_details(web3, token1_address)
    
        return PairDetails(
            pool,
            token0,
            token1,
            reverse_token_order=reverse_token_order,
        )


from .conversion import Conversion
from .log_context import LogContext
from .log_result import LogResult
from .chain_reorganization_detection import ChainReorganizationDetected
from .rpc_reorganization_monitor import JSONRPCReorganizationMonitor



# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

from web3 import Web3
from ..enums.nets_enum import NetsEnum
from ..enums.rpcs_enum import RPCEnum

DEFAULT_CHAIN = NetsEnum.LOCALHOST

class ConnectW3():

    def __init__(self, chain = None):        
        self.__chain = DEFAULT_CHAIN if chain == None else chain
        self.__w3 = None
     
    def apply(self):
        rpc = RPCEnum.get_rpc(self.__chain) 
        self.__w3 = Web3(Web3.HTTPProvider(rpc))

    def is_connect(self):
        return self.__w3.is_connected() if self.__w3 != None else False

    def get_w3(self):
        return self.__w3

    def get_chain_name(self):
        return self.__chain


# Copyright 2023–2025 Ian Moore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from web3 import Web3
from .event import Event
from ..data.filter import Filter
from .tools.conversion import Conversion
from ..utils.connect import ConnectW3
from ..enums.contracts_enum import JSONContractsEnum as JSONContracts
import math

class BurnEvent(Event):

    def __init__(self, connect_w3: ConnectW3):
        self.__connect_w3 = connect_w3   

    def record(self, event, abi_load):

        event_record = {}
        contract_type = abi_load.get_contract_name()

        match contract_type:
            case JSONContracts.IUniswapV2Pair:
                event_record = self._uni_v2_record(event, abi_load)
            case JSONContracts.UniswapV2Pair:
                event_record = self._uni_v2_record(event, abi_load)                
            case JSONContracts.UniswapV3Pool:
                event_record = self._uni_v3_record(event, abi_load)                
           
        return event_record    
                        
    def _uni_v2_record(self, event, abi_load):

        topics = event["topics"]
        arguments = Conversion().decode_data(event["data"])        

        chain_nm = self.__connect_w3.get_chain_name()
        contract_nm = abi_load.get_contract_name()
        platform_nm = abi_load.get_platform_name()
        
        event_record = {}
        event_record['chain'] = chain_nm
        event_record['contract'] = contract_nm.lower()
        event_record['type'] = event["event"].event_name.lower()
        event_record['platform'] = platform_nm
        event_record['address'] = event["address"]
        event_record['tx_hash'] = event["transactionHash"]
        event_record['blk_num'] = event["blockNumber"]
        event_record['timestamp'] = event["timestamp"]
        event_record['details'] = {}
        event_record['details']['web3_type'] = event["event"]
        event_record['details']['token0'] = Conversion().convert_uint256_hex_string_to_address(topics[1])
        event_record['details']['token1'] = Conversion().convert_uint256_hex_string_to_address(topics[2])
    
        if(len(arguments) >= 1):
            event_record['details']['amount0'] = Conversion().convert_int256_bytes_to_int(arguments[0])
        else:
            event_record['details']['amount0'] = math.nan
    
        if(len(arguments) >= 2):
            event_record['details']['amount1'] = Conversion().convert_int256_bytes_to_int(arguments[1])
        else:
            event_record['details']['amount1'] = math.nan   

        return event_record

    def _uni_v3_record(self, event, abi_load):

        event_signature, owner, tick_lower, tick_upper = event["topics"]
        args = Conversion().decode_data(event["data"])

        chain_nm = self.__connect_w3.get_chain_name()
        contract_nm = abi_load.get_contract_name()
        platform_nm = abi_load.get_platform_name()        
        
        event_record = {}
        event_record['chain'] = chain_nm
        event_record['contract'] = contract_nm.lower()
        event_record['type'] = event["event"].event_name.lower()
        event_record['platform'] = platform_nm
        event_record['address'] = event["address"]
        event_record['tx_hash'] = event["transactionHash"]
        event_record['blk_num'] = event["blockNumber"]
        event_record['timestamp'] = event["timestamp"]
        event_record['details'] = {}
        event_record['details']['web3_type'] = event["event"]
        event_record['details']['owner'] = Conversion().convert_uint256_hex_string_to_address(owner)
        event_record['details']['tick_lower'] = Conversion().convert_uint256_string_to_int(tick_lower, signed=True)
        event_record['details']['tick_upper'] = Conversion().convert_uint256_string_to_int(tick_upper, signed=True)    
        if(len(args) == 4):
            event_record['details']['liquidity_amount'] = Conversion().convert_int256_bytes_to_int(args[1])
            event_record['details']['amount0'] = Conversion().convert_int256_bytes_to_int(args[2])
            event_record['details']['amount1'] = Conversion().convert_int256_bytes_to_int(args[3])

        return event_record      
     
    def filter(self, contract, addr = None, fromBlock = None, toBlock = None):
        event_filt = contract.events.Burn.create_filter(fromBlock=fromBlock, toBlock=toBlock)
        return event_filt


# Copyright 2023–2025 Ian Moore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from .nets_enum import NetsEnum

DEFAULT_NET = NetsEnum.POLYGON

class RPCEnum:
    
    def get_key(net = DEFAULT_NET):
        match net:
            case NetsEnum.POLYGON:
                select_key = 'JSON_RPC_POLYGON'
            case NetsEnum.LOCALHOST:
                select_key = 'JSON_RPC_LOCALHOST'                
           
        return select_key 

    def get_rpc(net = DEFAULT_NET):
        match net:
            case NetsEnum.POLYGON:
                select_rpc = 'https://polygon-rpc.com'
            case NetsEnum.LOCALHOST:
                select_rpc = 'http://127.0.0.1:8545'  
            case _: 
                select_rpc = net 
           
        return select_rpc 


# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

"""Raw log event data conversion helpers."""

from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3

class Conversion:

    def __init__(self):
        pass    

    def decode_data(self, data: str) -> list[bytes]:
        """Split data of a log to uin256 results"""
    
        # {'address': '0x5c69bee701ef814a2b6a3edd4b1652cb9cc5aa6f', 'blockHash': '0x359d1dc4f14f9a07cba3ae8416958978ce98f78ad7b8d505925dad9722081f04', 'blockNumber': '0x98b723', 'data': '0x000000000000000000000000b4e16d0168e52d35cacd2c6185b44281ec28c9dc0000000000000000000000000000000000000000000000000000000000000001', 'logIndex': '0x22', 'removed': False, 'topics': ['0x0d3648bd0f6ba80134a33ba9275ac585d9d315f0ad8355cddefde31afa28d0e9', '0x000000000000000000000000a0b86991c6218b36c1d19d4a2e9eb0ce3606eb48', '0x000000000000000000000000c02aaa39b223fe8d0a0e5c4f27ead9083c756cc2'], 'transactionHash': '0xd07cbde817318492092cc7a27b3064a69bd893c01cb593d6029683ffd290ab3a', 'transactionIndex': '0x26', 'event': <class 'web3._utils.datatypes.PairCreated'>, 'timestamp': 1588710145}
        b = bytes.fromhex(data[2:])
        entries = []
        for i in range(0, len(b), 32):
            entries.append(b[i : i + 32])
        return entries
    
    
    def convert_uint256_bytes_to_address(self, raw: bytes | HexBytes) -> ChecksumAddress:
        """Convert raw uin256 from log data to addresses.
    
        .. note ::
    
            Ethereum address checksum might have a speed penalty for
            high speed operations.
    
        :param raw:
            Raw uint256 byte blob
    
        :return:
            Checksummed Ethereum address
        """
        assert type(raw) in (bytes, HexBytes), f"Received: {type(raw)}"
        assert len(raw) == 32
        return Web3.to_checksum_address(raw[12:])


    def convert_hex_bytes_to_string(self, raw: bytes | HexBytes) -> str:
        """Convert hex bytes from log data to string.
    
        :param hex:
            Hex string byte blob
    
        :return:
            Checksummed Ethereum address
        """
        if isinstance(raw, HexBytes):
            return raw.hex()  # or data.decode() if it's UTF-8 encoded
        elif isinstance(data, bytes):
            return raw.hex()
        elif isinstance(data, str):
            return raw
        else:
            raise TypeError("Unsupported data type")
    
    
    def convert_uint256_hex_string_to_address(self, hex: str) -> ChecksumAddress:
        """Convert raw uin256 from log data to address.
    
        :param hex:
            Hex string byte blob
    
        :return:
            Checksummed Ethereum address
        """
        assert type(hex) == str, f"Received: {type(hex)}"
        raw = HexBytes(hex)
        assert len(raw) == 32
        return Web3.to_checksum_address(raw[12:])
    
    
    def convert_int256_bytes_to_int(self, bytes32: bytes, *, signed: bool = False) -> int:
        """Convert raw bytes32 from log data to addresses ints.
    
        :param signed:
            Default to unsigned uint256. Set true for int256.
        """
        return int.from_bytes(bytes32, "big", signed=signed)
    
    
    def convert_uint256_string_to_address(self, bytes32: str) -> ChecksumAddress:
        """Convert raw uint256 from log data to address.
    
        .. note ::
    
            Ethereum address checksum might have a speed penalty for
            high speed operations.
    
        :param bytes32:
            E.g. `0x00000000000000000000000006af07097c9eeb7fd685c692751d5c66db49c215`
        """
        assert bytes32.startswith("0x")
        raw = bytes.fromhex(bytes32[2:])
        assert len(raw) == 32
        return Web3.to_checksum_address(raw[12:])
    
    
    def convert_uint256_string_to_int(self, bytes32: str, *, signed: bool = False) -> int:
        """Convert raw uint256 from log data to int.
    
        :param bytes32:
            E.g. `0x00000000000000000000000006af07097c9eeb7fd685c692751d5c66db49c215`
        """
        assert bytes32.startswith("0x")
        raw = bytes.fromhex(bytes32[2:])
        return int.from_bytes(raw, "big", signed=signed)
    
    
    def convert_jsonrpc_value_to_int(self, val: str | int) -> int:
        """Convert hex string or int to int.
    
        Depending on the used JSON-RPC node,
        they may return hex encoded values or JSON numbers
        in JSON-RPC type. We need to be able to support both node and
        do the compatibility hack here.
        """
    
        if type(val) == int:
            # EthereumTester
            return val
    
        # Hex number
        return int(val, 16)



# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

"""Bunch of random utilities."""
import calendar
import datetime
import logging
import os
import random
import socket
import time
from typing import Optional, Tuple
from urllib.parse import urlparse

import psutil


logger = logging.getLogger(__name__)

#: Ethereum 0x0000000000000000000000000000000000000000 address as a string
ZERO_ADDRESS_STR = "0x0000000000000000000000000000000000000000"    

class BaseUtils():

    def __init__(self):
        pass        

    def sanitise_string(self, s: str) -> str:
        """Remove null characters."""
        # https://stackoverflow.com/a/18762899/315168
        return s.replace("\x00", "\U0000FFFD")
    
    
    def is_localhost_port_listening(self, port: int, host="localhost") -> bool:
        """Check if a localhost is running a server already.
    
        See https://www.adamsmith.haus/python/answers/how-to-check-if-a-network-port-is-open-in-python
    
        :return: True if there is a process occupying the port
        """
    
        a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        location = (host, port)
        result_of_check = a_socket.connect_ex(location)
        return result_of_check == 0
    
    
    def find_free_port(self, min_port: int = 20_000, max_port: int = 40_000, max_attempt: int = 20) -> int:
        """Find a free localhost port to bind.
    
        Does by random.
    
        .. note ::
    
            Subject to race condition, but should be rareish.
    
        :param min_port:
            Minimum port range
    
        :param max_port:
            Maximum port range
    
        :param max_attempt:
            Give up and die with an exception if no port found after this many attempts.
    
        :return:
            Free port number
        """
    
        assert type(min_port) == int
        assert type(max_port) == int
        assert type(max_attempt) == int
    
        for attempt in range(0, max_attempt):
            random_port = random.randrange(start=min_port, stop=max_port)
            logger.info("Attempting to allocate port %d to Anvil", random_port)
            if not is_localhost_port_listening(random_port, "127.0.0.1"):
                return random_port
    
        raise RuntimeError(f"Could not open a port with a spec: {min_port} - {max_port}, {max_attempt} attempts")
    
    
    def shutdown_hard(
        self,
        process: psutil.Popen,
        log_level: Optional[int] = None,
        block=True,
        block_timeout=30,
        check_port: Optional[int] = None,
    ) -> Tuple[bytes, bytes]:
        """Kill Psutil process.
    
        - Straight out OS `SIGKILL` a process
    
        - Log output if necessary
    
        - Use port listening to check that the process goes down
          and frees its ports
    
        :param process:
            Process to kill
    
        :param block:
            Block the execution until the process has terminated.
    
            You must give `check_port` option to ensure we enforce the shutdown.
    
        :param block_timeout:
            How long we give for process to clean up after itself
    
        :param log_level:
            If set, dump anything in Anvil stdout to the Python logging using level `INFO`.
    
        :param check_port:
            Check that TCP/IP localhost port is freed after shutdown
    
        :return:
            stdout, stderr as string
        """
    
        stdout = b""
        stderr = b""
    
        if process.poll() is None:
            # Still alive, we need to kill to read the output
            process.kill()
    
        for line in process.stdout.readlines():
            stdout += line
            if log_level is not None:
                logger._log(log_level, "stdout: %s", line.decode("utf-8").strip())
    
        for line in process.stderr.readlines():
            stderr += line
            if log_level is not None:
                logger._log(log_level, "stderr: %s", line.decode("utf-8").strip())
    
        if block:
            assert check_port is not None, "Give check_port to block the execution"
            deadline = time.time() + 30
            while time.time() < deadline:
                if not is_localhost_port_listening(check_port):
                    # Port released, assume Anvil/Ganache is gone
                    return stdout, stderr
    
            raise AssertionError(f"Could not terminate Anvil in {block_timeout} seconds, stdout is %d bytes, stderr is %d bytes", len(stdout), len(stderr))
    
        return stdout, stderr
    
    
    def to_unix_timestamp(self, dt: datetime.datetime) -> float:
        """Convert Python UTC datetime to UNIX seconds since epoch.
    
        Example:
    
        .. code-block:: python
    
            import datetime
            from eth_defi.utils import to_unix_timestamp
    
            dt = datetime.datetime(1970, 1, 1)
            unix_time = to_unix_timestamp(dt)
            assert unix_time == 0
    
        :param dt:
            Python datetime to convert
    
        :return:
            Datetime as seconds since 1970-1-1
        """
        # https://stackoverflow.com/a/5499906/315168
        return calendar.timegm(dt.utctimetuple())
    
    
    def get_url_domain(self, url: str) -> str:
        """Redact URL so that only domain is displayed.
    
        Some services e.g. infura use path as an API key.
        """
        parsed = urlparse(url)
        if parsed.port in (80, 443, None):
            return parsed.hostname
        else:
            return f"{parsed.hostname}:{parsed.port}"
    
    
    def setup_console_logging(self):
        """Set up coloured log output.
    
        - Helper function to have nicer logging output in tutorial scripts.
        - Tune down some noisy dependency library logging
        """
    
        try:
            import coloredlogs
        except ImportError as e:
            raise RuntimeError("coloredlogs package missing - please install with pip first before running") from e
    
        level = os.environ.get("LOG_LEVEL", "info").upper()
    
        fmt = "%(asctime)s %(name)-44s %(message)s"
        date_fmt = "%H:%M:%S"
        coloredlogs.install(level=level, fmt=fmt, date_fmt=date_fmt)
    
        logging.basicConfig(level=level, handlers=[logging.StreamHandler()])
    
        # Mute noise
        logging.getLogger("web3.providers.HTTPProvider").setLevel(logging.WARNING)
        logging.getLogger("web3.RequestManager").setLevel(logging.WARNING)
        logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)



# ─────────────────────────────────────────────────────────────────────────────
# Apache 2.0 License (DeFiPy)
# ─────────────────────────────────────────────────────────────────────────────
# Copyright 2023–2025 Ian Moore
# Email: defipy.devs@gmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from web3 import Web3
from uniswappy.erc import ERC20
from ...abi.abi_load import ABILoad
from ...enums.platforms_enum import PlatformsEnum
from ...enums.contracts_enum import JSONContractsEnum

class FetchToken:
    
    def __init__(self, w3):
        self.w3 = w3

    def apply(self, token_address):
        name = self.get_token_symbol(token_address)
        decimal = self.get_token_decimals(token_address)
        return ERC20(name, token_address, decimal)

    def amt_to_decimal(self, tkn, amt):
        return amt/(10**tkn.token_decimal)
        
    def get_erc20_abi(self, token_address):
        abi_obj = ABILoad(PlatformsEnum.ERC, JSONContractsEnum.ERC20)  # Load ABI here  
        contract_instance = abi_obj.apply(self.w3, token_address)
        return contract_instance.abi
        
    def get_token_symbol(self, token_address):
        """Fetch the token symbol for a given ERC-20 token address."""
        # Minimal ERC-20 ABI for symbol()
        erc20_abi = self.get_erc20_abi(token_address)
        try:
            # Ensure address is checksummed
            checksum_address = self.w3.to_checksum_address(token_address)
            contract = self.w3.eth.contract(address=checksum_address, abi=erc20_abi)
            symbol = contract.functions.symbol().call()
            return symbol
        except Exception as e:
            print(f"Error fetching symbol for {token_address}: {e}")
            return None

    def get_token_name(self, token_address):
        """Fetch the token symbol for a given ERC-20 token address."""
        # Minimal ERC-20 ABI for symbol()
        erc20_abi = self.get_erc20_abi(token_address)
        try:
            # Ensure address is checksummed
            checksum_address = self.w3.to_checksum_address(token_address)
            contract = self.w3.eth.contract(address=checksum_address, abi=erc20_abi)
            name = contract.functions.name().call()
            return name
        except Exception as e:
            print(f"Error fetching symbol for {self.token_address}: {e}")
            return None
    
    def get_token_supply(self, token_address):
        """Fetch the token symbol for a given ERC-20 token address."""
        # Minimal ERC-20 ABI for symbol()
        erc20_abi = self.get_erc20_abi(token_address)
        try:
            # Ensure address is checksummed
            checksum_address = self.w3.to_checksum_address(token_address)
            contract = self.w3.eth.contract(address=checksum_address, abi=erc20_abi)
            supply = contract.functions.totalSupply().call()
            return supply
        except Exception as e:
            print(f"Error fetching symbol for {self.token_address}: {e}")
            return None
    
    def get_token_decimals(self, token_address):
        """Fetch the number of decimals for a given ERC-20 token address."""
        erc20_abi = self.get_erc20_abi(token_address)
        try:
            checksum_address = self.w3.to_checksum_address(token_address)
            contract = self.w3.eth.contract(address=checksum_address, abi=erc20_abi)
            decimals = contract.functions.decimals().call()
            return decimals
        except Exception as e:
            print(f"Error fetching decimals for {token_address}: {e}")
            return None
    



# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

"""In-house filter implementation.

Wrap low-level JSON-RPC filters with more manageable Python code.
"""

from dataclasses import dataclass
from typing import Dict, Optional, List, Type

from eth_bloom import BloomFilter
from web3.contract.contract import ContractEvent


@dataclass
class Filter:
    """Internal filter used to match events.

    A helper class to deal with `eth_getLogs` JSON API.

    This can be used with

    - Historical events: :py:mod:`eth_defi.event_reader.reader`

    - Live events: :py:mod:`eth_defi.event_reader.websocket`
    """

    #: Preconstructed topic hash -> Event mapping
    topics: Dict[str, ContractEvent]

    #: Bloom filter to match block headers
    #: TODO: Currently unsupported
    bloom: Optional[BloomFilter]

    #: Get events from a single contract only.
    #:
    #: For multiple contracts give a list of addresses.
    contract_address: Optional[str | List[str]] = None

    @staticmethod
    def create_filter(address: Optional[str | List[str]], event_types: List[Type[ContractEvent]]) -> "Filter":
        topics = {event_type.build_filter().topics[0]: event_type for event_type in event_types}

        filter = Filter(
            contract_address=address,
            bloom=None,
            topics=topics,
        )

        return filter



# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0


from typing import Dict, TypeAlias, Union
from eth_typing import HexAddress
from web3.contract import Contract
from web3 import Web3
from ..abi.abi_load import ABILoad

#: Manage internal registry of deployed contracts
#:
#: Lower case address -> Contract mapping.
ContractRegistry: TypeAlias = Dict[str, Contract]


class ContractDeploymentFailed(Exception):
    """Did not get successful tx receipt from a deployment."""

    def __init__(self, tx_hash, msg):
        super().__init__(msg)
        self.tx_hash = tx_hash

class Deploy():

    def __init__(self):
        pass
    
    def deploy_contract(
        self,
        web3: Web3,
        contract: Union[str, Contract],
        deployer: str,
        *constructor_args,
        register_for_tracing=True,
    ) -> Contract:
        """Deploys a new contract from ABI file.
    
        A generic helper function to deploy any contract.
    
        Example:
    
        .. code-block:: python
    
            token = deploy_contract(web3, deployer, "ERC20Mock.json", name, symbol, supply)
            print(f"Deployed ERC-20 token at {token.address}")
    
        If you need to verify the deployed contract use :py:func:`eth_defi.foundry.forge.deploy_contract_with_forge`.
    
        :param web3:
            Web3 instance
    
        :param contract:
            Contract file path as string or contract proxy class
    
        :param deployer:
            Deployer account
    
        :param constructor_args:
            Other arguments to pass to the contract's constructor
    
        :param register_for_tracing:
            Make the symbolic contract information available on web3 instance.
    
            See :py:func:`get_contract_registry`
    
        :raise ContractDeploymentFailed:
            In the case we could not deploy the contract.
    
        :return:
            Contract proxy instance
    
        """
        if isinstance(contract, str):
            Contract = ABILoading().get_contract(web3, contract)
    
            # Used in trace.py
            contract_name = contract.replace(".json", "")
    
        else:
            Contract = contract
            contract_name = None
    
        tx_hash = Contract.constructor(*constructor_args).transact({"from": deployer})
    
        tx_receipt = web3.eth.wait_for_transaction_receipt(tx_hash)
        if tx_receipt["status"] != 1:
            raise ContractDeploymentFailed(tx_hash, f"Contract {contract_name} deployment failed with args {constructor_args}, tx hash is {tx_hash.hex()}")
    
        instance = Contract(address=tx_receipt["contractAddress"])
    
        if register_for_tracing:
            instance.name = contract_name
            self.register_contract(web3, tx_receipt["contractAddress"], instance)
    
        return instance


    def register_contract(self, web3, address: HexAddress, instance: Contract):
        """Register a contract for tracing.
    
        See :py:func:`deploy_contract`.
        """
        assert type(address) == str, f"address is {type(address)}, expected str"
        registry = self.get_or_create_contract_registry(web3)
        registry[address.lower()] = instance

    def get_registered_contract(self, web3, address: str) -> Contract:
        """Get a contract that was deployed with the registry.
    
        - Resolve a symbolic contract information based on the contract address and our contract registry
    
        - See :py:func:`eth_defi.deploy.deploy_contract` how to deploy a registered contract
    
        Example:
    
        .. code-block:: python
    
             from eth_defi.deploy import get_registered_contract
    
             contract = get_registered_contract(web3, "0x1613beb3b2c4f22ee086b2b38c1476a3ce7f78e8")
             assert contract.name == "VaultSpecificGenericAdapter"
    
        :param address:
            Contract address as a hex string
    
        :return:
            The known Contract instance at the registry or `None` if the contract was not registered/deployed through registry mechanism.
        """
        assert type(address) == str
        registry = self.get_or_create_contract_registry(web3)
        return registry.get(address.lower())    


    def get_or_create_contract_registry(self, web3: Web3) -> ContractRegistry:
        """Get a contract registry associated with a Web3 connection.
    
        - Only relevant for test sessions
    
        - Assumes one web3 instance per test
    
        - Useful to make traces symbolic in :py:mod:`eth_defi.trace`
    
        :param web3:
            Web3 test session
    
        :return:
            Mapping of address -> deployed contract instance
        """
        if not hasattr(web3, "contract_registry"):
            web3.contract_registry = {}
    
        return web3.contract_registry






# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0


from ...abi.abi_load import ABILoad
from ...enums.init_event_enum import InitEventEnum as InitEvent
from ...utils.connect import ConnectW3
from ..event import Event
from ..tools.log_result import LogResult
from ..tools.conversion import Conversion
from ..tools.rpc_reorganization_monitor import JSONRPCReorganizationMonitor
from .read_events import ReadEvents
import pandas as pd

class RetrieveEvents:

    def __init__(self, connect: ConnectW3, abi: ABILoad, verbose = True):
        self.__connect = connect   
        self.__abi = abi 
        self.__w3 = self.__connect.get_w3()
        self.__contract = None
        self.__verbose = verbose
        self.__w3.middleware_onion.clear()
    
    def apply(self, event_type, address = None, start_block = None, end_block = None):

        assert self.__connect.is_connect(), 'WEB3SCOUT Event Reader: NOT_CONNECTED'
        assert address != None, 'WEB3SCOUT Event Reader: NO_ADDRESS'

        self.__contract = self.retrieve_contract(address)
        event = InitEvent().apply(self.__connect, event_type)
        read_events = self.gen_read_events(event, start_block, end_block)
        return self.to_dict(read_events) 

    def get_contract(self):
        return self.__contract

    def to_dict(self, read_events = {}):
        
        dict_events = {}
        for k, evt in enumerate(read_events):
            evt_record = self.reorg_event_record(evt)
            dict_events[k] = evt_record
        return dict_events    
    
    def to_dataframe(self, dict_events):
        return pd.DataFrame.from_dict(dict_events, orient='index')

    def gen_read_events(self, event, start_block = None, end_block = None):
        s_block = 1 if start_block == None else start_block
        e_block = self.latest_block() if end_block == None else end_block
        event_filt = event.filter(self.__contract, fromBlock=s_block, toBlock=e_block)
        read_events = event_filt.get_all_entries() 
        return read_events

    def reorg_event_record(self,evt):
        event_fields = ['blockNumber','event','address','blockHash','logIndex','transactionHash','transactionIndex','args']
        event_record = {}
        for field in event_fields:

            if(field == 'blockHash' or field == 'transactionHash'):
                event_record[field] = Conversion().convert_hex_bytes_to_string(evt[field])
            else:
                event_record[field] = evt[field]
                
        return event_record

    def retrieve_contract(self, address):
        chksum_addr = address if address == None else self.__w3.to_checksum_address(address)
        return self.__abi.apply(self.__w3, chksum_addr)
    
    def latest_block(self):
        reorg_mon = JSONRPCReorganizationMonitor(self.__w3, check_depth=3)
        reorg_mon.load_initial_block_headers(block_count=5)
        return reorg_mon.get_last_block_live()


from .event import Event
from .burn_event import BurnEvent
from .mint_event import MintEvent
from .swap_event import SwapEvent
#from .sync_event import SyncEvent
#from .create_event import CreateEvent
from .transfer_event import TransferEvent



# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

from dataclasses import dataclass
from typing import Tuple

@dataclass(slots=True, frozen=True)
class ChainReorganizationResolution:
    """How did we fare getting hashes and timestamps for the latest blocks."""

    #: What we know is the chain tip on our node
    #:
    #: This is the latest block at the JSON-RPC node.
    #: We can read data up to this block.
    last_live_block: int

    #: What we know is the block for which we do not need to perform rollback
    #:
    #: This is the block number that does not need to purged from your internal database.
    #: All previously read events that have higher block number should be purged.
    #:
    latest_block_with_good_data: int

    #: Did we detect any reorgs in this chycle
    reorg_detected: bool

    def __repr__(self):
        return f"<reorg:{self.reorg_detected} last_live_block: {self.last_live_block:,}, latest_block_with_good_data:{self.latest_block_with_good_data:,}>"

    def get_read_range(self) -> Tuple[int, int]:
        """Get the range of blocks we should read on this poll cycle.

        - This range may overlap your previous event read range.

        - You should discard any data that's older than the start of the range

        - You should be prepared to read an event again

        :return:
            (start block, end block) inclusive range

        """
        assert self.last_live_block >= self.latest_block_with_good_data, f"Last block from the node: {self.last_live_block}, last block we have read: {self.latest_block_with_good_data}"
        return (self.latest_block_with_good_data + 1, self.last_live_block)



# Copyright 2023–2025 Ian Moore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from .event_type_enum import EventTypeEnum as EventType
from ..event.mint_event import MintEvent
from ..event.swap_event import SwapEvent
from ..event.burn_event import BurnEvent
from ..event.create_event import CreateEvent
from ..event.transfer_event import TransferEvent
from ..event.sync_event import SyncEvent

DEFAULT_EVENT = EventType.MINT

class InitEventEnum:
    
    def apply(self, connect, event_type = DEFAULT_EVENT):
        match event_type:
            case EventType.MINT:
                event = MintEvent(connect)
            case EventType.SWAP:
                event = SwapEvent(connect)     
            case EventType.BURN:
                event = BurnEvent(connect)         
            case EventType.SYNC:
                event = SyncEvent(connect)      
            case EventType.TRANSFER:
                event = TransferEvent(connect)     
            case EventType.CREATE:
                event = CreateEvent(connect)                    
                           
           
        return event 


# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

from web3.contract.contract import ContractEvent
from typing import TypedDict, Optional, List
from .log_context import LogContext

class LogResult(TypedDict):
    """One emitted Solidity event.

    - Type mappings for a raw Python :py:class:`dict` object

    - Designed for high performance at the cost of readability and usability

    - The values are untranslated hex strings to maximize the reading speed of events

    - See :py:func:`decode_event` how to turn to ABI converted data

    - See :py:mod:`eth_defi.event_reader.reader` for more information

    Example data (PancakeSwap swap):

    .. code-block:: text

        {
            'address': '0xc91cd2b9c9aafe494cf3ccc8bee7795deb17231a',
            'blockHash': '0x3bc60abea8fca30516f48b0374542b9c8fa554061c8802d7bcd4211fffbf6caf',
            'blockNumber': 30237147,
            'chunk_id': 30237147,
            'context': None,
            'data': '0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000de90d34e1f2e65c0000000000000000000000000000000000000000000018e627902bfb974416f90000000000000000000000000000000000000000000000000000000000000000',
            'event': <class 'web3._utils.datatypes.Swap'>,
            'logIndex': '0x3',
            'removed': False,
            'timestamp': 1690184818,
            'topics': ['0xd78ad95fa46c994b6551d0da85fc275fe613ce37657fb8d5e3d130840159d822',
                    '0x00000000000000000000000064c425b6fa04873ea460cda049b340d79cf859d7',
                    '0x000000000000000000000000ad1fedfb04377c4b849cef6ef9627bca41955fa0'],
            'transactionHash': '0xf2287653559f01d8afba9ae00386d453b731699b784851f7a8504d41dee7503b',
            'transactionIndex': '0x1'
        }

    """

    #: User passed context for the event reader
    context: LogContext

    #: Contract event matches for this raw log
    #:
    #: To use web3.py helpers to decode this log.
    #:
    #: This event instance is just a class reference and does
    #: not contain any bound data.
    #:
    event: ContractEvent

    #: Smart contract address
    address: str

    #: Block where the event was
    blockHash: str

    #: Block number as hex string
    blockNumber: str

    #: UNIX timestamp of the block number.
    #: Synthesized by block reader code, not present in the receipt.
    #: May be None if timestamp fetching is disabled for the speed reasons.
    timestamp: Optional[int]

    #: Transaction where the event occred
    transactionHash: str

    #: Log index as a hex number
    logIndex: str

    #: Topics in this receipt.
    #: `topics[0]` is always the event signature.
    #:
    #: TODO: Whether these are strings or HexBytes depends on the EVM backend and Web3 version.
    #: Resolve this so that results are normalised to one type.
    #:
    #: See :py:mod:`eth_defi.reader.conversion` how to get Python values out of this.
    #:
    topics: List[str]

    #: Block reorg helper
    removed: bool

    #: Data related to the event
    #:
    #: As raw hex dump from the JSON-RPC.
    #:
    #: See :py:func:`eth_defi.reader.conversion.decode_data` to split to args.
    #:
    data: str


# Copyright 2023–2025 Ian Moore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



from ..abi.abi_load import ABILoad
from ..utils.connect import ConnectW3

class ViewContract():

    def __init__(self, connect: ConnectW3, abi: ABILoad, verbose = False):
        self.__connect = connect   
        self.__abi = abi 
        self.__w3 = self.__connect.get_w3()
        self.__contract = None
        self.__verbose = verbose

    def apply(self, address):
        contract_interface = self.get_contract_interface()
        contract = self.__w3.eth.contract(address=address, abi=contract_interface['abi'])
        view_fns = self.retrieve_view_funcs()

        view_fns_res = {}
        for k, fn_str in enumerate(view_fns) :
            try:
                fn_res = self._call_fn(contract, fn_str)
                fn_str_pout = self._str_output(f'[{k}]', fn_str+'()', max_len = 5)
                res = self._str_output(fn_str_pout, fn_res)
                view_fns_res[fn_str] = fn_res
                if(self.__verbose): print(res)
            except ValueError:
                print('function call error')

        return view_fns_res

    
    def retrieve_view_funcs(self):

        contract_interface = self.get_contract_interface()
        
        view_fns = []; c = 0
        for k, record in enumerate(contract_interface['abi']):
            if(record['type'] == 'function' and len(record['inputs']) == 0 and record['stateMutability'] == 'view'):
                c+=1
                if(self.__verbose): print(f"[{c}] View function: {record['name']}(), inputs: {record['inputs']}")
                view_fns.append(record['name'])   

        return view_fns    

    def get_contract_interface(self):
        fname = self.__abi.get_abi_path() 
        return self.__abi.get_abi_by_filename(fname)    


    def _concat_str(self, str0, max_len = None):
        max_len = 40 if max_len == None else max_len
        str0 = "".join([str0, " "])
        str_len = len(str0)
        return str0.ljust(max_len, ' ')
    
    def _str_output(self, in0, in1, max_len = None):    
        print_str_list = [self._concat_str(in0, max_len), str(in1)]    
        return ''.join(print_str_list)
    
    def _call_fn(self, contract, str_fn):
        attr = getattr(contract.functions, str_fn)
        return attr().call()
    


# SPDX-Anchored-Identifier: ICMOORE-2025-DEFIPY
# This file participates in a symbolic cognition substrate.

from .abi.abi_load import ABILoad
from .event.tools.chain_reorganization_detection import ChainReorganizationDetected
from .event.tools.rpc_reorganization_monitor import JSONRPCReorganizationMonitor
from .event.tools.conversion import Conversion
from .event.tools.log_context import LogContext
from .event.tools.log_result import LogResult
from .event.process.read_events import ReadEvents
from .event.process.retrieve_events import RetrieveEvents
from .event.mint_event import MintEvent
from .event.swap_event import SwapEvent
from .event.sync_event import SyncEvent
from .event.transfer_event import TransferEvent
from .event.create_event import CreateEvent
from .event.burn_event import BurnEvent
from .data.filter import Filter
from .data.reorganization_monitor import ReorganizationMonitor
from .data.chain_reorganization_resolution import ChainReorganizationResolution
from .data.pair import PairDetails
from .utils.progress_update import ProgressUpdate
from .utils.base_utils import BaseUtils
from .contract.deploy import Deploy
from .contract.view import ViewContract
from .uniswap_v2.fetch_pair_details import FetchPairDetails
from .token.token import Token
from .token.fetch.fetch_token import FetchToken
from .utils.connect import ConnectW3
from .enums.event_type_enum import EventTypeEnum as EventType
from .enums.init_event_enum import InitEventEnum as InitEvent
from .enums.nets_enum import NetsEnum as Net
from .enums.rpcs_enum import RPCEnum as RPC
from .enums.platforms_enum import PlatformsEnum as Platform
from .enums.contracts_enum import JSONContractsEnum as JSONContract





# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

import logging
from dataclasses import dataclass, field
from typing import Dict, Optional, Type, Callable, Tuple, Iterable
from abc import *
from tqdm import tqdm
from .chain_reorganization_resolution import ChainReorganizationResolution
from ..event.tools.chain_reorganization_detection import ChainReorganizationDetected
import pandas as pd


logger = logging.getLogger(__name__)

from eth_defi.event_reader.block_header import BlockHeader, Timestamp

@dataclass()
class ReorganizationMonitor(ABC):
    """Watch blockchain for reorgs.

    Most EMV blockchains have several minor chain organisations per day,
    when your node switched from one chain tip to another, due to
    block propagation issues. Any application reading blockchain
    event data must be able to detect such reorganisations
    and purge incorrect data from their data feeds.

    - Abstract base class for different ways
      to support chain reorganisations

    - Maintain the state where our blockchain read cursor is,
      using :py:meth:`get_last_block_read`

    - Ingest and maintain the state of the last read blocks
      using :py:meth:`update_chain`

    - Check block headers for chain reorganisations when
      reading events from the chain using :py:meth:`check_block_reorg`

    - Manages the service for block timestamp lookups,
      :py:meth:`get_block_timestamp`

    - Save and load block header state to disk cache,
      because APIs are slow, using :py:meth:`load_pandas`
      and :py:meth:`to_pandas`

    Example:

    .. code-block:: python

        import os
        import time

        from web3 import HTTPProvider, Web3

        from eth_defi.abi import get_contract
        from eth_defi.chain import install_chain_middleware
        from eth_defi.event_reader.filter import Filter
        from eth_defi.event_reader.reader import read_events, LogResult,
        from eth_defi.event_reader.reorganisation_monitor import JSONRPCReorganisationMonitor


        def main():

            json_rpc_url = os.environ.get("JSON_RPC_POLYGON", "https://polygon-rpc.com")
            web3 = Web3(HTTPProvider(json_rpc_url))
            web3.middleware_onion.clear()
            install_chain_middleware(web3)

            # Get contracts
            Pair = get_contract(web3, "sushi/UniswapV2Pair.json")

            filter = Filter.create_filter(
                address=None,  # Listen events from any smart contract
                event_types=[Pair.events.Swap]
            )

            reorg_mon = JSONRPCReorganisationMonitor(web3, check_depth=3)

            reorg_mon.load_initial_block_headers(block_count=5)

            processed_events = set()

            latest_block = None

            # Keep reading events as they land
            while True:
                chain_reorg_resolution = reorg_mon.update_chain()
                start, end = chain_reorg_resolution.get_read_range()

                if chain_reorg_resolution.reorg_detected:
                    print("Chain reorg warning")

                evt: LogResult
                for evt in read_events(
                    web3,
                    start_block=start,
                    end_block=end,
                    filter=filter,
                ):
                    # How to uniquely identify EVM logs
                    key = evt["blockHash"] + evt["transactionHash"] + evt["logIndex"]

                    # The reader may cause duplicate events as the chain tip reorganises
                    if key not in processed_events:
                        print(f"Swap at block {evt['blockNumber']:,} tx: {evt['transactionHash']}")
                        processed_events.add(key)

                if end != latest_block:
                    print(f"Latest block is {end:,}")
                    latest_block = end

                time.sleep(0.5)


        if __name__ == "__main__":
            main()


    """

    #: Internal buffer of our block data
    #:
    #: Block number -> Block header data
    block_map: Dict[int, BlockHeader] = field(default_factory=dict)

    #: Last block served by :py:meth:`update_chain` in the duty cycle
    last_block_read: int = 0

    #: How many blocks we replay from the blockchain to detect any chain organisations
    #:
    #: Done by :py:meth:`figure_reorganisation_and_new_blocks`.
    #: Adjust this for your EVM chain.
    check_depth: int = 20

    #: How many times we try to re-read data from the blockchain in the case of reorganisation.
    #:
    #: If our node constantly feeds us changing data give up.
    max_cycle_tries = 10

    #: How long we allow our node to catch up in the case there has been a change in the chain tip.
    #:
    #: If our node constantly feeds us changing data give up.
    reorg_wait_seconds = 5

    def has_data(self) -> bool:
        """Do we have any data available yet."""
        return len(self.block_map) > 0

    def get_last_block_read(self) -> int:
        """Get the number of the last block served by update_chain()."""
        return self.last_block_read

    def get_block_by_number(self, block_number: int) -> BlockHeader:
        """Get block header data for a specific block number from our memory buffer."""
        return self.block_map.get(block_number)

    def skip_to_block(self, block_number: int):
        """Skip scanning initial chain and directly start from a certain block."""
        assert type(block_number) == int, f"Got: {block_number}"
        logger.info(f"{self}: skipping to block {block_number:,}")
        self.last_block_read = block_number

    def load_initial_block_headers(self, block_count: Optional[int] = None, start_block: Optional[int] = None, tqdm: Optional[Type[tqdm]] = None, save_callable: Optional[Callable] = None) -> Tuple[int, int]:
        """Get the initial block buffer filled up.

        You can call this during the application start up,
        or when you start the chain. This interface is designed
        to keep the application on hold until new blocks have been served.

        :param block_count:
            How many latest block to load

            Give `start_block` or `block_count`.

        :param start_block:
            What is the first block to read.

            Give `start_block` or `block_count`.

        :param tqdm:
            To display a progress bar

        :param save_callable:
            Save after every block.

            Called after every block.

            TODO: Hack. Design a better interface.

        :return:
            The initial block range to start to work with
        """

        end_block = self.get_last_block_live()

        if block_count:
            assert not start_block, "Give block_cout or start_block"
            start_block = max(end_block - block_count, 1)
        else:
            pass

        if len(self.block_map) > 0:
            # We have some initial data from the last (aborted) run,
            # We always need to start from the last save because no gaps in data allowed
            oldest_saved_block = max(self.block_map.keys())
            start_block = oldest_saved_block + 1

        blocks = end_block - start_block

        if tqdm:
            progress_bar = tqdm(total=blocks, colour="green")
            progress_bar.set_description(f"Downloading block headers {start_block:,} - {end_block:,}")
        else:
            progress_bar = None

        last_saved_block = None
        for block in self.fetch_block_data(start_block, end_block):
            self.add_block(block)

            if save_callable:
                last_saved_block, _ = save_callable()
                if last_saved_block:
                    last_saved_block_str = f"{last_saved_block:,}" if last_saved_block else "-"
                    progress_bar.set_postfix({"Last saved block": last_saved_block_str}, refresh=False)

            if progress_bar:
                progress_bar.update(1)

        if progress_bar:
            progress_bar.close()

        return start_block, end_block

    def add_block(self, record: BlockHeader):
        """Add new block to header tracking.

        Blocks must be added in order.
        """

        assert isinstance(record, BlockHeader)

        block_number = record.block_number
        assert block_number not in self.block_map, f"Block already added: {block_number}"
        self.block_map[block_number] = record

        if self.last_block_read != 0:
            assert self.last_block_read == block_number - 1, f"Blocks must be added in order. Last block we have: {self.last_block_read}, the new record is: {record}"
        self.last_block_read = block_number

    def check_block_reorg(self, block_number: int, block_hash: str) -> Optional[Timestamp]:
        """Check that newly read block matches our record.

        - Called during the event reader

        - Event reader gets the block number and hash with the event

        - We have initial `block_map` in memory, previously buffered in

        - We check if any of the blocks in the block map have different values
          on our event produces -> in this case we know there has been a chain reorganisation

        If we do not have records, ignore.

        :raise ChainReorganizationDetected:
            When any if the block data in our internal buffer
            does not match those provided by events.
        """
        original_block = self.block_map.get(block_number)
        if original_block is not None:
            if original_block.block_hash != block_hash:
                raise ChainReorganizationDetected(block_number, original_block.block_hash, block_hash)

            return original_block.timestamp

        return None

    def truncate(self, latest_good_block: int):
        """Delete data after a block number because chain reorg happened.

        :param latest_good_block:
            Delete all data starting after this block (exclusive)
        """
        assert self.last_block_read
        for block_to_delete in range(latest_good_block + 1, self.last_block_read + 1):
            del self.block_map[block_to_delete]
        self.last_block_read = latest_good_block

    def figure_reorganisation_and_new_blocks(self, max_range: Optional[int] = 1_000_000):
        """Compare the local block database against the live data from chain.

        Spot the differences in (block number, block header) tuples
        and determine a chain reorg.

        :param max_range:
            Abort if we need to scan more than this amount of blocks.

            This is because giving too long block range to scan is likely to
            take forever on non-graphql nodes.

            Set `None` to ignore.

        :raise ChainReorganizationDetected:
            When any if the block data in our internal buffer
            does not match those provided by events.
        """
        chain_last_block = self.get_last_block_live()
        check_start_at = max(self.last_block_read - self.check_depth, 1)

        logger.info(f"figure_reorganisation_and_new_blocks(), range {check_start_at:,} - {chain_last_block:,}, last block we have is {self.last_block_read:,}, check depth is %d", self.check_depth)

        if max_range is not None:
            range_len = chain_last_block - check_start_at
            if range_len > max_range:
                raise TooLongRange(f"Attempt to scan too long block range. {check_start_at:,} - {chain_last_block:,}. Max range: {max_range:,}.\nFor long scan ranges, please pass a flag to ignore.")

        for block in self.fetch_block_data(check_start_at, chain_last_block):
            self.check_block_reorg(block.block_number, block.block_hash)
            if block.block_number not in self.block_map:
                self.add_block(block)

    def get_block_timestamp(self, block_number: int) -> int:
        """Return UNIX UTC timestamp of a block."""

        if not self.block_map:
            raise BlockNotAvailable("We have no records of any blocks")

        if block_number not in self.block_map:
            last_recorded_block_num = max(self.block_map.keys())
            raise BlockNotAvailable(f"Block {block_number} has not data, the latest live block is {self.get_last_block_live()}, last recorded is {last_recorded_block_num}")

        return self.block_map[block_number].timestamp

    def get_block_timestamp_as_pandas(self, block_number: int) -> pd.Timestamp:
        """Return UNIX UTC timestamp of a block."""

        ts = self.get_block_timestamp(block_number)
        return pd.Timestamp.utcfromtimestamp(ts).tz_localize(None)

    def update_chain(self) -> ChainReorganizationResolution:
        """Update the internal memory buffer of block headers from the blockchain node.

        - Do several attempt to read data (as a fork can cause other forks can cause fork)

        - Give up after some time if we detect the chain to be in a doom loop

        :return:
            What block range the consumer application should read.

            What we think about the chain state.
        """

        tries_left = self.max_cycle_tries
        max_purge = self.get_last_block_read()
        reorg_detected = False
        while tries_left > 0:
            try:
                self.figure_reorganisation_and_new_blocks()
                return ChainReorganizationResolution(self.last_block_read, max_purge, reorg_detected=reorg_detected)
            except ChainReorganizationDetected as e:
                logger.info("Chain reorganisation detected: %s", e)

                latest_good_block = e.block_number - 1

                reorg_detected = True

                if max_purge:
                    max_purge = min(latest_good_block, max_purge)
                else:
                    max_purge = e.block_number

                self.truncate(latest_good_block)
                tries_left -= 1
                time.sleep(self.reorg_wait_seconds)

        raise ReorganisationResolutionFailure(f"Gave up chain reorg resolution. Last block: {self.last_block_read}, attempts {self.max_cycle_tries}")

    def to_pandas(self, partition_size: int = 0) -> pd.DataFrame:
        """Convert the data to Pandas DataFrame format for storing.

        :param partition_size:
            To partition the outgoing data.

            Set 0 to ignore.

        """
        data = [asdict(h) for h in self.block_map.values()]
        return BlockHeader.to_pandas(data, partition_size)

    def load_pandas(self, df: pd.DataFrame):
        """Load block header data from Pandas data frame.

        :param df:

            Pandas DataFrame exported with :py:meth:`to_pandas`.
        """
        block_map = BlockHeader.from_pandas(df)
        self.restore(block_map)

    def restore(self, block_map: dict):
        """Restore the chain state from a saved data.

        :param block_map:
            Block number -> Block header dictionary
        """
        assert type(block_map) == dict, f"Got: {type(block_map)}"
        self.block_map = block_map
        self.last_block_read = max(block_map.keys())

    @abstractmethod
    def fetch_block_data(self, start_block, end_block) -> Iterable[BlockHeader]:
        """Read the new block headers.

        :param start_block:
            The first block where to read (inclusive)

        :param end_block:
            The block where to read (inclusive)
        """

    @abstractmethod
    def get_last_block_live(self) -> int:
        """Get last block number"""


# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

from typing import Optional
from decimal import Decimal
from dataclasses import dataclass
from functools import cached_property
from web3.contract import Contract
from eth_typing import HexAddress

@dataclass
class TokenDetails:
    """ERC-20 token Python presentation.

    - A helper class to work with ERC-20 tokens.

    - Read on-chain data, deal with token value decimal conversions.

    - Any field can be ``None`` for non-well-formed tokens.

    Example how to get USDC details on Polygon:

    .. code-block:: python

        usdc = fetch_erc20_details(web3, "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174")  # USDC on Polygon
        formatted = f"Token {usdc.name} ({usdc.symbol}) at {usdc.address} on chain {usdc.chain_id}"
        assert formatted == "Token USD Coin (PoS) (USDC) at 0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174 on chain 137"
    """

    #: The underlying ERC-20 contract proxy class instance
    contract: Contract

    #: Token name e.g. ``USD Circle``
    name: Optional[str] = None

    #: Token symbol e.g. ``USDC``
    symbol: Optional[str] = None

    #: Token supply as raw units
    total_supply: Optional[int] = None

    #: Number of decimals
    decimals: Optional[int] = None

    def __eq__(self, other):
        """Token is the same if it's on the same chain and has the same contract address."""
        assert isinstance(other, TokenDetails)
        return (self.contract.address == other.contract.address) and (self.chain_id == other.chain_id)

    def __hash__(self):
        """Token hash."""
        return hash((self.chain_id, self.contract.address))

    def __repr__(self):
        return f"<{self.name} ({self.symbol}) at {self.contract.address}, {self.decimals} decimals, on chain {self.chain_id}>"

    @cached_property
    def chain_id(self) -> int:
        """The EVM chain id where this token lives."""
        return self.contract.w3.eth.chain_id

    @property
    def address(self) -> HexAddress:
        """The address of this token."""
        return self.contract.address

    def convert_to_decimals(self, raw_amount: int) -> Decimal:
        """Convert raw token units to decimals.

        Example:

        .. code-block:: python

            details = fetch_erc20_details(web3, token_address)
            # Convert 1 wei units to edcimals
            assert details.convert_to_decimals(1) == Decimal("0.0000000000000001")

        """
        return Decimal(raw_amount) / Decimal(10**self.decimals)

    def convert_to_raw(self, decimal_amount: Decimal) -> int:
        """Convert decimalised token amount to raw uint256.

        Example:

        .. code-block:: python

            details = fetch_erc20_details(web3, token_address)
            # Convert 1.0 USDC to raw unit with 6 decimals
            assert details.convert_to_raw(1) == 1_000_000

        """
        return int(decimal_amount * 10**self.decimals)

    def fetch_balance_of(self, address: HexAddress | str, block_identifier="latest") -> Decimal:
        """Get an address token balance.

        :param block_identifier:
            A specific block to query if doing archive node historical queries

        :return:
            Converted to decimal using :py:meth:`convert_to_decimal`
        """
        raw_amount = self.contract.functions.balanceOf(address).call(block_identifier=block_identifier)
        return self.convert_to_decimals(raw_amount)

    @staticmethod
    def generate_cache_key(chain_id: int, address: str) -> int:
        """Generate a cache key for this token.

        - Cached by (chain, address) tuple

        - Validate the inputs before generating the key
        """
        assert type(chain_id) == int
        assert type(address) == str
        assert address.startswith("0x")
        return hash((chain_id, address.lower()))


# Copyright 2023–2025 Ian Moore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

@dataclass(frozen=True)
class JSONContractsEnum:
    ERC20: str = "ERC20MockDecimals"
    IUniswapV2Pair: str = "IUniswapV2Pair"
    UniswapV2Pair: str = "UniswapV2Pair"
    UniswapV2Router02: str = "UniswapV2Router02"
    UniswapV2Factory: str = "UniswapV2Factory"
    UniswapV3Pool: str = "UniswapV3Pool"
    UniswapV3Factory: str = "UniswapV3Factory"
    UniV2IndexedYieldLinearExitPool: str = "UniV2IndexedYieldLinearExitPool"
    MintableRCIndexedYieldLinearExitPoolStudy: str = "MintableRCIndexedYieldLinearExitPoolStudy"






from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setup(name='Web3Scout',
      version='0.0.7',
      description='Onchain Event Framework for DeFiPy',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/defipy-devs/web3scout',
      author = "icmoore",
      author_email = "defipy.devs@gmail.com",
      license="Apache-2.0",
      classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
            "Intended Audience :: Developers",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
      ],
        package_dir = {"web3scout": "python/prod"},
        packages=[
            "web3scout",
            "web3scout.event",
            "web3scout.event.tools",
            "web3scout.event.process",
            "web3scout.abi",
            "web3scout.utils",
            "web3scout.data",
            "web3scout.enums",
            "web3scout.token",
            "web3scout.token.fetch",
            "web3scout.contract",
            "web3scout.uniswap_v2"
        ],
        install_requires=['web3', 
                          'eth_abi', 
                          'eth_typing',
                          'eth_tester',
                          'eth_bloom',
                          'eth_utils', 
                          'web3-ethereum-defi',
                          'hexbytes', 
                          'pandas',
                          'defipy >= 1.0.8'
                         ],      
        include_package_data=True,
        zip_safe=False,
    )



# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

import json
import re
from functools import lru_cache
from pathlib import Path
from typing import Optional, Sequence, Type, Union

import eth_abi
from eth_abi import decode
from eth_typing import HexAddress
from eth_utils import encode_hex, function_abi_to_4byte_selector
from eth_utils.abi import _abi_to_signature, function_signature_to_4byte_selector
from hexbytes import HexBytes
from web3 import Web3
from web3._utils.abi import get_abi_input_names, get_abi_input_types
from web3._utils.contracts import encode_abi, get_function_info
from web3.contract.contract import Contract, ContractFunction

# Cache loaded ABI files in-process memory for speedup
from web3.datastructures import AttributeDict

from ..utils.base_utils import ZERO_ADDRESS_STR

from ..enums.platforms_enum import PlatformsEnum
from ..enums.contracts_enum import JSONContractsEnum

DEFAULT_CONTRACT = JSONContractsEnum.IUniswapV2Pair
DEFAULT_PLATFORM = PlatformsEnum.SUSHI

# How big are our ABI and contract caches
_CACHE_SIZE = 512

class ABILoad:

    def __init__(self, platform = None, contract = None):
        self.__platform_name = DEFAULT_PLATFORM if platform == None else platform
        self.__contract_name = DEFAULT_CONTRACT if contract == None else contract
        self.__abi_path = self.__platform_name + '/' + self.__contract_name + '.json'

    def apply(self, web3: Web3, address: Optional[str] = None):
        return self.get_contract(web3, self.__abi_path, address)  

    def get_contract_name(self):
        return self.__contract_name

    def get_platform_name(self):
        return self.__platform_name    

    def get_abi_path(self):
        return self.__abi_path      

    @lru_cache(maxsize=_CACHE_SIZE)
    def get_abi_by_filename(self, fname: str) -> dict:
        """Reads a embedded ABI file and returns it.
    
        Example::
    
            abi = get_abi_by_filename("ERC20Mock.json")
    
        You are most likely interested in the keys `abi` and `bytecode` of the JSON file.
    
        Loaded ABI files are cache in in-process memory to speed up future loading.
    
        Any results are cached.
    
        :param web3: Web3 instance
        :param fname: `JSON filename from supported contract lists <https://github.com/tradingstrategy-ai/web3-ethereum-defi/tree/master/eth_defi/abi>`_.
        :return: Full contract interface, including `bytecode`.
        """
    
        here = Path(__file__).resolve().parent        
        abi_path = here / Path(fname)
        abi_path = str(abi_path).replace("abi/","configs/")
        with open(abi_path, "rt", encoding="utf-8") as f:
            abi = json.load(f)
        return abi
    
    
    @lru_cache(maxsize=_CACHE_SIZE)
    def get_contract(
        self, 
        web3: Web3,
        fname: str | Path,
        address: Optional[str] = None,
        bytecode: Optional[str] = None,
    ) -> Type[Contract]:
        """Create a Contract proxy class from our bundled contracts or filesystem..
    
        `See Web3.py documentation on Contract instances <https://web3py.readthedocs.io/en/stable/contracts.html#contract-deployment-example>`_.
    
        Any results are cached. Web3 connection is part of the cache key.
    
        .. note ::
    
            This function cannot do linking. See :py:func:`get_linked_contract`
            if the bytecode contains link markers.
    
        Example:
    
            pass
    
        :param web3:
            Web3 instance
    
        :param bytecode:
            Override bytecode payload for the contract
    
        :param fname:
            Solidity compiler artifact.
    
            Use slash prefixed path for absolute lookups.
    
            `JSON filename from supported contract lists <https://github.com/tradingstrategy-ai/web3-ethereum-defi/tree/master/eth_defi/abi>`_.
    
        :return:
            Contract proxy class
        """
    
        contract_interface = self.get_abi_by_filename(fname)
        abi = contract_interface["abi"]
        
        if address is None and bytecode is None:
            bytecode = contract_interface["bytecode"]
    
            if type(bytecode) == dict:
                # Sol 0.8 / Forge?
                # Contains keys object, sourceMap, linkReferences
                bytecode = bytecode["object"]
            else:
                # Sol 0.6 / legacy
                # Bytecode hex is directly in the key.
                pass
                
            contract = web3.eth.contract(abi=abi, bytecode=bytecode)  
            
        elif address is not None and bytecode is None:
            contract = web3.eth.contract(address=address, abi=abi)
            

        
        return contract
    
    
    def get_linked_contract(
        self, 
        web3: Web3,
        fname: str | Path,
        hardhat_export_data: Optional[dict] = None,
    ) -> Type[Contract]:
        """Create a Contract proxy class from our bundled contracts or filesystem and links it Solidity bytecode.
    
        Needed when contracts contain references to libraries. The contract bytecode
        must be processed and placeholders must be replaced by the on-chain addresses
        of the deployed library contracts.
    
        Example:
    
        .. code-block:: python
    
            path = self.path.joinpath("artifacts/@aave/core-v3/contracts/mocks/tokens/MintableERC20.sol/MintableERC20.json")
            return get_linked_contract(web3, path, get_aave_hardhard_export())
    
        .. note ::
    
            If you do not need linking use :py:func:`get_contract` which is faster.
    
        :param web3:
            Web3 instance
    
        :param fname:
            Solidity compiler artifact.
    
            Use slash prefixed path for absolute lookups.
    
            `JSON filename from supported contract lists <https://github.com/tradingstrategy-ai/web3-ethereum-defi/tree/master/eth_defi/abi>`_.
    
        :param hardhat_export_data:
            Hardhat deployment export data to link bytecode.
    
            A JSON file generated by `hardhat deploy --export` command.
    
        :return:
            Contract proxy class
        """
    
        contract_interface = self.get_abi_by_filename(fname)
        abi = contract_interface["abi"]
        bytecode = contract_interface["deployedBytecode"]
    
        link_references = contract_interface["linkReferences"]
        bytecode = self.link_libraries_hardhat(bytecode, link_references, hardhat_export_data)
    
        Contract = web3.eth.contract(abi=abi, bytecode=bytecode)
        return Contract
    
    
    def get_deployed_contract(
        self, 
        web3: Web3,
        fname: str | Path,
        address: Union[HexAddress, str],
        register_for_tracing: bool = True,
    ) -> Contract:
        """Get a Contract proxy objec for a contract deployed at a specific address.
    
        `See Web3.py documentation on Contract instances <https://web3py.readthedocs.io/en/stable/contracts.html#contract-deployment-example>`_.
    
        :param web3:
            Web3 instance
    
        :param fname:
            `JSON filename from supported contract lists <https://github.com/tradingstrategy-ai/web3-ethereum-defi/tree/master/eth_defi/abi>`_.
    
        :param address:
            Ethereum address of the deployed contract
    
        :param register_for_tracing:
            Add the contract to the deployment registry if not already there.
    
        :return:
            `web3.contract.Contract` proxy
        """
        assert address
    
        address = Web3.to_checksum_address(address)
    
        Contract = self.get_contract(web3, fname)
        contract = Contract(address)
    
        if register_for_tracing:
            # TODO: Currently hack around circular imports, move functoins
            from pachira.contract.deploy import Deploy
    
            registered_contract = Deploy().get_registered_contract(web3, address)
            if registered_contract is None:
                Deploy().register_contract(web3, address, contract)
    
        return contract
    
    
    def get_transaction_data_field(self, tx: AttributeDict) -> str:
        """Get the "Data" payload of a transaction.
    
        Ethereum Tester has this in tx.data while Ganache has this in tx.input.
        Yes, it is madness.
    
        Example:
    
        .. code-block::
    
            tx = web3.eth.get_transaction(tx_hash)
            function, input_args = router.decode_function_input(get_transaction_data_field(tx))
            print("Transaction {tx_hash} called function {function}")
    
        """
        if "data" in tx:
            return tx["data"]
        else:
            return tx["input"]
    
    
    def encode_with_signature(self, function_signature: str, args: Sequence) -> bytes:
        """Mimic Solidity's abi.encodeWithSignature() in Python.
    
        This is a Python equivalent for `abi.encodeWithSignature()`.
    
        Example:
    
        .. code-block:: python
    
                payload = encode_with_signature("init(address)", [my_address])
                assert type(payload) == bytes
    
        :param function_signature:
            Solidity function signature that can be hashed to a selector.
    
            ABI fill be extractd from this signature.
    
        :param args:
            Argument values to be encoded.
        """
    
        assert type(args) in (tuple, list)
    
        function_selector = Web3.keccak(text=function_signature)
        selector_text = self.function_signature[function_signature.find("(") + 1 : function_signature.rfind(")")]
        arg_types = selector_text.split(",")
        encoded_args = eth_abi.encode(arg_types, args)
        return function_selector + encoded_args
    
    
    def encode_function_args(self, func: ContractFunction, args: Sequence) -> bytes:
        """Mimic Solidity's abi.encodeWithSignature() in Python.
    
        Uses `web3.Contract.functions` prepared function as the ABI source.
    
        :param func:
            Function which arguments we are going to encode.
    
        :param args:
            Argument values to be encoded.
        """
        assert isinstance(func, ContractFunction)
    
        web3 = func.w3
    
        fn_abi, fn_selector, aligned_fn_arguments = get_function_info(
            func.fn_name,
            web3.codec,
            func.contract_abi,
            args=args,
        )
        arg_types = [t["type"] for t in fn_abi["inputs"]]
        encoded_args = eth_abi.encode(arg_types, args)
        return encoded_args
    
    
    def encode_function_call(
        self, 
        func: ContractFunction,
        args: Sequence,
    ) -> HexBytes:
        """Encode function selector + its arguments as data payload.
    
        Uses `web3.Contract.functions` prepared function as the ABI source.
    
        See also :py:func:`encode_function_args`.
    
        :param func:
            Function which arguments we are going to encode.
    
        :param args:
            Argument values to be encoded.
    
        :return:
            Solidity's function selector + argument payload.
    
        """
        w3 = func.w3
        contract_abi = func.contract_abi
        fn_abi = func.abi
        fn_identifier = func.function_identifier
        fn_abi, fn_selector, fn_arguments = get_function_info(
            # type ignored b/c fn_id here is always str b/c FallbackFn is handled above
            fn_identifier,  # type: ignore
            w3.codec,
            contract_abi,
            fn_abi,
            args,
        )
        encoded = encode_abi(w3, fn_abi, fn_arguments, fn_selector)
        return HexBytes(encoded)
    
    
    def decode_function_args(
        self, 
        func: ContractFunction,
        data: bytes | HexBytes,
    ) -> dict:
        """Decode binary CALL or CALLDATA to a Solidity function,
    
        Uses `web3.Contract.functions` prepared function as the ABI source.
    
        :param func:
            Function which arguments we are going to encode.
    
        :param data:
            Extracted from a transaction data field or EVM memoryo trace.
    
        :return:
            Ordered dict of the decoded arguments
        """
        assert isinstance(func, ContractFunction)
        fn_abi = func.abi
    
        arg_names = get_abi_input_names(fn_abi)
        arg_types = get_abi_input_types(fn_abi)
        arg_tuple = decode(arg_types, data)
    
        return dict(zip(arg_names, arg_tuple))
    
    
    def humanise_decoded_arg_data(self, args: dict) -> dict:
        """Make decoded arguments more human readable.
    
        - All arguments are converted to good text types
    
        See :py:func:`decode_function_args`
    
        :return:
            Ordered dict of decoded arguments, easier to read
        """
    
        def _humanize(v):
            if type(v) == bytes:
                return v.hex()
            return v
    
        return {k: _humanize(v) for k, v in args.items()}
    
    
    def link_libraries_hardhat(self, bytecode: str, link_references: dict, hardhat_export: dict):
        """Link Solidity libraries based on Hardhat deployment.
    
        .. warning ::
    
            Preliminary implementation
    
        See :py:func:`get_linked_contract` for details.
    
        :param bytecode:
            Raw bytecode of a Solidity contract.
    
            Get from ABI file.
    
            Bytecode must be a in string format, because placeholders are not parseable hex.
    
        :param link_references:
            List of binary sequences we need to replaced by a contract filename.
    
            Get from ABI file.
    
        :param hardhat_export:
            Hardhat's export format.
    
            You get with `hardhat deploy --export`.
    
        :return:
            Linked bytecode
        """
    
        assert type(bytecode) == str, f"Got {type(bytecode)}"
    
        assert bytecode.startswith("0x")
    
        hex_blob = bytecode[2:]
    
        # Remove placeholders and replace them with zeroes,
        # so that we can convert the bytecode to binary
        # https://stackoverflow.com/a/16160048/315168
        zeroes = str(ZERO_ADDRESS_STR)[2:]
        fixed_hex_blob = re.sub(r"__\$(.*?)\$__", zeroes, hex_blob, flags=re.DOTALL)
    
        data = bytearray.fromhex(fixed_hex_blob)
    
        def _get_contract_address(name: str):
            contracts = hardhat_export["contracts"]
            contract = contracts.get(name)
            assert contract, f"Hardhat export does not contain a contract entry for {name}"
            return contract["address"]
    
        for ref_file, ref_data in link_references.items():
            # 'contracts/protocol/libraries/logic/BorrowLogic.sol': {'BorrowLogic': [{'length': 20, 'start': 4405}, {'length': 20, 'start': 5081}, {'length': 20, 'start': 7441}, {'length': 20, 'start': 7604}, {'length': 20, 'start': 10111}, {'length': 20, 'start': 12083}]},
            for contract_name, ref_array in ref_data.items():
                address = _get_contract_address(contract_name)
                byte_address = bytes.fromhex(address[2:])
                for ref in ref_array:
                    start = ref["start"]
                    length = ref["length"]
                    data[start : start + length] = byte_address
                    # print(f"Linking {contract_name} {start} {address}")
    
        return data
    
    
    def get_function_selector(self, func: ContractFunction) -> bytes:
        """Get Solidity function selector.
    
        Does not support multiple Solidity functions with the same name, but
        different arguments. On multiple functions
        use one first declared in ABI.
    
        Example:
    
        .. code-block:: python
    
            selector = get_function_selector(uniswap_v2.router.functions.swapExactTokensForTokens)
            assert selector.hex() == 38ed1739
    
        :param func:
            Unbound or bound contract function proxy
    
        :return:
            Solidity function selector.
    
            First 32-bit (4 bytes) keccak hash.
        """
    
        contract_abi = func.contract_abi
        # https://stackoverflow.com/a/8534381/315168
        fn_abi = next((a for a in contract_abi if a.get("name") == func.fn_name), None)
        assert fn_abi, f"Could not find function {func.fn_name} in Contract ABI"
        function_signature = _abi_to_signature(fn_abi)
        fn_selector = function_signature_to_4byte_selector(function_signature)  # type: ignore
        return fn_selector



# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0

class LogContext:
    """An abstract context class you can pass around for the log results.

    Subclass this and add your own data / methods.

    See `scripts/read-uniswap-v2-pairs-and-swaps.py` for an example.
    """


# This file contains code adapted from web3-ethereum-defi (https://github.com/Kartograf/web3-ethereum-defi)
# Licensed under the MIT License.
# Original copyright (c) 2023 Kartograf contributors.

# Additional modifications by Ian Moore, 2023–2025
# Licensed under the Apache License, Version 2.0


"""Uniswap v2 pair info.

- See :py:class:`PairDetails` for a helper class
  to get price and other data from the trading pairs of :Uniswap v2 like DEXes
"""

from dataclasses import dataclass
from decimal import Decimal
from typing import Union, Optional

from eth_typing import HexAddress
from web3.contract import Contract

from ..abi.abi_load import ABILoad
from ..data.token_details import TokenDetails
from ..token.token import Token


@dataclass(frozen=True, slots=True)
class PairDetails:
    """Uniswap v2 trading pair info.

    An example usage how to get the latest price of a pair on PancakeSwap.
    The `PairDetails` class will do an automatic conversion of prices to human-readable, decimal format:

    .. code-block:: python

        from web3 import Web3, HTTPProvider

        from eth_defi.chain import install_chain_middleware
        from eth_defi.uniswap_v2.pair import fetch_pair_details

        web3 = Web3(HTTPProvider("https://bsc-dataseed.bnbchain.org"))

        print(f"Connected to chain {web3.eth.chain_id}")

        # BNB Chain does its own things
        install_chain_middleware(web3)

        # Find pair addresses on TradingStrategy.ai
        # https://tradingstrategy.ai/trading-view/binance/pancakeswap-v2/bnb-busd
        pair_address = "0x58f876857a02d6762e0101bb5c46a8c1ed44dc16"
        wbnb = "0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c"
        busd = "0xe9e7CEA3DedcA5984780Bafc599bD69ADd087D56"

        # PancakeSwap has this low level encoded token0/token1 as BNB/BUSD
        # in human-readable token order
        # and we do not need to swap around
        reverse_token_order = False

        pair = fetch_pair_details(
            web3,
            pair_address,
            reverse_token_order,
        )

        assert pair.token0.address == wbnb
        assert pair.token1.address == busd

        price = pair.get_current_mid_price()

        # Assume 1 BUSD = 1 USD
        print(f"The current price of PancakeSwap BNB/BUSD is {price:.4f} USD")

    """

    #: Pool contract
    #:
    #: https://docs.uniswap.org/contracts/v2/reference/smart-contracts/pair#getreserves
    contract: Contract

    #: One pair of tokens
    token0: TokenDetails

    #: One pair of tokens
    token1: TokenDetails

    #: Store the human readable token order on this data.
    #:
    #: If false then pair reads as token0 symbol (WETH) - token1 symbol (USDC).
    #:
    #: If true then pair reads as token1 symbol (USDC) - token0 symbol (WETH).
    reverse_token_order: Optional[bool] = None

    def __eq__(self, other):
        """Implemented for set()"""
        assert isinstance(other, PairDetails)
        return self.address == other.address

    def __hash__(self) -> int:
        """Implemented for set()"""
        return int(self.address, 16)

    def __repr__(self):
        return f"<Pair {self.get_base_token().symbol}-{self.get_quote_token().symbol} at {self.address}>"

    @property
    def address(self) -> HexAddress:
        """Get pair contract address"""
        return self.contract.address

    @property
    def checksum_free_address(self) -> str:
        """Get pair contract address, all lowercase."""
        return self.contract.address.lower()

    def get_base_token(self) -> TokenDetails:
        """Get human-ordered base token."""
        assert self.reverse_token_order is not None, "Reverse token order flag must be check before this operation is possible"
        if self.reverse_token_order:
            return self.token1
        else:
            return self.token0

    def get_quote_token(self) -> TokenDetails:
        """Get human-ordered quote token."""
        assert self.reverse_token_order is not None, "Reverse token order flag must be check before this operation is possible"
        if self.reverse_token_order:
            return self.token0
        else:
            return self.token1

    def convert_price_to_human(self, reserve0: int, reserve1: int, reverse_token_order=None):
        """Convert the price obtained through Sync event

        :param reverse_token_order:
            Decide token order for human (base, quote token) order.
            If set, assume quote token is token0.

            IF set to None, use value from the data.

        """

        if reverse_token_order is None:
            reverse_token_order = self.reverse_token_order

        if reverse_token_order is None:
            reverse_token_order = False

        token0_amount = self.token0.convert_to_decimals(reserve0)
        token1_amount = self.token1.convert_to_decimals(reserve1)

        if reverse_token_order:
            return token0_amount / token1_amount
        else:
            return token1_amount / token0_amount

    def get_current_mid_price(self) -> Decimal:
        """Return the price in this pool.

        Calls `getReserves()` over JSON-RPC and calculate
        the current price basede on the pair reserves.

        See https://docs.uniswap.org/contracts/v2/reference/smart-contracts/pair#getreserves

        :return:
            Quote token / base token price in human digestible form
        """
        assert self.reverse_token_order is not None, "Reverse token order must be set to get the natural price"
        reserve0, reserve1, timestamp = self.contract.functions.getReserves().call()
        return self.convert_price_to_human(reserve0, reserve1, self.reverse_token_order)



